/*
    *************************************************************************************

    ZipWriter v1.2
    (c) Jarmo Pietiläinen 2013-2014
    http://z0b.kapsi.fi

    This software is provided 'as-is', without any express or implied warranty.
    In no event will the authors be held liable for any damages arising from the
    use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it freely,
    subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not
         be misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    *************************************************************************************
*/

// --------------------------------------------------------------------------------------
// All APPNOTE.TXT references are to the version 6.3.3, revised on September 1 2012.
// http://www.pkware.com/documents/casestudies/APPNOTE.TXT
// --------------------------------------------------------------------------------------

// detect the compile-time platform
#if defined(_WIN32) || defined(WIN32) || defined(__CYGWIN__) || defined(__MINGW32__) || defined(__BORLANDC__)
    #define ZIPWRITER_WINDOWS

    // make _wfopen() work
    #undef __STRICT_ANSI__
#endif

#include "zipwriter.h"

#ifdef ZIPWRITER_WINDOWS
    // need this for some Windows-specific functions
    #include <windows.h>
#else
    // generic Unix/Linux stuff
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <unistd.h>
    #include <ctime>
#endif

#include <string>
#include <vector>
#include <cstdio>
#include <stdint.h>

#include <zlib.h>

using namespace std;

// various record headers
enum : uint32_t {
    kHeader_LocalFile = 0x04034B50,
    kHeader_CD = 0x02014B50,
    kHeader_CDEnd = 0x06054B50,
    //kHeader_CDEnd64 = 0x06064B50,
    //kHeader_CDLocator64 = 0x07064B50,
};

// some values for the "version needed to extract" field in the file header
enum : uint16_t {
    kVersionNeeded_Default = 10,
    kVersionNeeded_Deflate = 20,
    kVersionNeeded_Directory = 20,
    //kVersionNeeded_ZIP64 = 45,
};

// supported compression methods
enum : uint16_t {
    kCompression_None = 0,
    kCompression_Deflate = 8
};

// some General Purpose Field flags
enum : uint16_t {
    kGPF_EFS = (1 << 11),           // UTF-8 encoded file name
};

// Maximum comment length. Applies to both file comments and the archive comment.
// There is no way around this limitation.
static const size_t kMaxCommentLength = 65535;

#ifdef ZIPWRITER_WINDOWS
    // I hate using pragmas to control the packing, but nothing else seems to
    // work reliably. -mno-ms-bitfields is a stupid compatibility hack and
    // I won't bother with compatibility hacks anymore.
    #define PACKED_STRUCT
    #pragma pack(push)
    #pragma pack(1)
#else
    // for non-broken compilers
    #define PACKED_STRUCT __attribute__((packed))
#endif

// end of central directory
// APPNOTE.TXT chapter 4.3.16
struct CDEnd_t {
    uint32_t id;                // kHeader_CDEnd
    uint16_t diskNum;
    uint16_t cdDisk;
    uint16_t filesOnDisk;
    uint16_t totalFiles;
    uint32_t size;
    uint32_t offset;
    uint16_t commentLen;
} PACKED_STRUCT;

#ifdef ZIPWRITER_WINDOWS
    #pragma pack(pop)
#endif

#undef PACKED_STRUCT

// verify the structure padding
static_assert(sizeof(CDEnd_t) == 22, "sizeof(CDEnd_t) != 22, check your compiler settings");

// The file header, stores everything related to a file. Contains the information for local
// file headers and central directory headers, in one structure. Some fields from the spec
// have been omitted (like diskette numbers), because we don't support all of those features.
// Combines parts from APPNOTE.TXT chapters 4.3.7 (local file) and 4.3.12 (central directory).
struct fileHeader_t {
    uint16_t versionNeeded;
    uint16_t versionMadeBy;         // set using compile-time directives
    uint16_t gpf;
    uint16_t method;
    uint16_t time;
    uint16_t date;
    uint32_t crc32;
    uint32_t csize;
    uint32_t osize;
    vector<uint8_t> name;           // UTF-8 encoded file name
    vector<uint8_t> comment;        // UTF-8 encoded file comment
    vector<uint8_t> extra;
    uint16_t internalAttrs;
    uint32_t externalAttrs;
    uint32_t offset;

    fileHeader_t() :
        versionNeeded(0),
        versionMadeBy(0),
        gpf(0),
        method(0),
        time(0),
        date(0),
        crc32(0),
        csize(0),
        osize(0),
        internalAttrs(0),
        externalAttrs(0),
        offset(0) { }

    ~fileHeader_t() { }
};

struct writerData_t {
    // output file handle
    FILE *out;

    // individual file headers
    std::vector<fileHeader_t> headers;

    // the main archive comment, see the manual for notes about character encodings
    std::string archiveComment;

    // global "version made by" field, set during compile-time and copied to
    // every file and directory that gets added in the archive
    uint16_t versionMadeBy;

    // current error code
    int errorCode;
};

// --------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------
// UTILITY FUNCTIONS

// STDIO file wrapper, closes the file automatically once it goes out of scope
class file {
private:
    FILE *f;

public:
    file() : f(nullptr) { }

    ~file()
    {
        close();
    }

#ifdef ZIPWRITER_WINDOWS
    bool open(const wchar_t *name, const wchar_t *mode)
#else
    bool open(const char *name, const char *mode)
#endif
    {
        if (!name || !mode)
            return false;

        // replace the existing handle only if we can successfully open the new file
#ifdef ZIPWRITER_WINDOWS
        FILE *n = _wfopen(name, mode);
#else
        FILE *n = fopen(name, mode);
#endif

        if (!n)
            return false;

        close();
        f = n;

        return true;
    }

    void close()
    {
        if (f) {
            fclose(f);
            f = nullptr;
        }
    }

    // transparent type conversion
    operator FILE *()
    {
        return f;
    }

    operator bool()
    {
        return !!f;
    }
};

static void write16(FILE *f, const uint16_t v)
{
    fwrite(&v, 1, sizeof(uint16_t), f);
}

static void write32(FILE *f, const uint32_t v)
{
    fwrite(&v, 1, sizeof(uint32_t), f);
}

static void writeVector8(FILE *f, const vector<uint8_t> &v)
{
    if (!v.empty())
        fwrite(&v[0], 1, v.size(), f);
}

// changes backward slashes to forward slashes
static void forwardSlashes(vector<uint8_t> &name)
{
    for (uint8_t &c : name)
        if (c == '\\')
            c = '/';
}

// --------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------
// PLATFORM-SPECIFIC BITS

#ifdef ZIPWRITER_WINDOWS

// Gets the "external" file infos, like the last modification time/date, attributes, size, etc.
static bool getExternalFileInfo(const wstring &name, fileHeader_t &header,
                                const bool isDir = false)
{
    DWORD attrs = FILE_ATTRIBUTE_NORMAL;

    if (isDir)
        attrs |= FILE_FLAG_BACKUP_SEMANTICS;

    HANDLE h = CreateFileW(name.c_str(), GENERIC_READ, FILE_SHARE_READ, nullptr,
                           OPEN_EXISTING, attrs, nullptr);

    if (h == INVALID_HANDLE_VALUE)
        return false;

    // get the file size
    DWORD size = GetFileSize(h, nullptr);

    if (size == INVALID_FILE_SIZE) {
        CloseHandle(h);
        return false;
    }

    header.osize = (uint32_t)size;

    // get the last modification date/time in DOS format
    FILETIME m, local;

    GetFileTime(h, nullptr, nullptr, &m);
    CloseHandle(h);

    FileTimeToLocalFileTime(&m, &local);
    FileTimeToDosDateTime(&local, &header.date, &header.time);

    // get the external attributes (I guess a better name would be "platform-specific attributes"?)
    DWORD ret = GetFileAttributesW(name.c_str());

    if (ret == INVALID_FILE_ATTRIBUTES)
        return false;

    header.externalAttrs = (uint32_t)ret;

    return true;
}

// Converts from UTF-8 to UCS-2. Windows only. Returns false if failed.
static bool UTF8ToUCS2(const string &utf8, wstring &ucs2)
{
    ucs2.clear();

    if (utf8.empty())
        return true;

    const int len = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, utf8.c_str(), utf8.length(), NULL, 0);

    if (len == 0)
        return false;

    ucs2.resize(len);

    if (!MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, utf8.c_str(), utf8.length(), &ucs2[0], len))
        return false;

    return true;
}

#else

// Gets the "external" file infos, like the last modification time/date, attributes, size, etc.
static bool getExternalFileInfo(const string &name, fileHeader_t &header,
                                const bool isDir = false)
{
    struct stat s;

    if (stat(name.c_str(), &s) == -1)
        return false;

    // get the last modification date/time in DOS format (subtract 80 from
    // the year, because that's the difference between 1900 and 1980 epochs)
    struct tm *tm = localtime(&s.st_mtime);

    header.date = ((tm->tm_year - 80) << 9) | ((tm->tm_mon + 1) << 5) | tm->tm_mday;
    header.time = (tm->tm_hour << 11) | (tm->tm_min << 5) | (tm->tm_sec / 2);

    // get the file size (dangerous integer truncation!)
    header.osize = (uint32_t)s.st_size;

    // get the external attributes (I guess a better name would be "platform-specific attributes"?)
    // @@@TODO: I'm not sure if this is the right way, zipinfo -v reports the correct values, but...
    header.externalAttrs = (uint32_t)s.st_mode << 16;

    return true;
}

#endif

// --------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------
// ZIPWRITER CLASS MEMBERS

// Can't use fwrite() here, because the same structure is shared between local and central
// file headers and fwrite() cannot differentiate between them.

// writes a local file header
static void writeLocalFileHeader(FILE *out, const fileHeader_t &hdr)
{
    write32(out, kHeader_LocalFile);
    write16(out, hdr.versionNeeded);
    write16(out, hdr.gpf);
    write16(out, hdr.method);
    write16(out, hdr.time);
    write16(out, hdr.date);
    write32(out, hdr.crc32);
    write32(out, hdr.csize);
    write32(out, hdr.osize);
    write16(out, hdr.name.size());
    write16(out, hdr.extra.size());
    writeVector8(out, hdr.name);
    writeVector8(out, hdr.extra);
}

// writes a central directory file header
static void writeCDFileHeader(FILE *out, const fileHeader_t &hdr)
{
    write32(out, kHeader_CD);
    write16(out, hdr.versionMadeBy);
    write16(out, hdr.versionNeeded);
    write16(out, hdr.gpf);
    write16(out, hdr.method);
    write16(out, hdr.time);
    write16(out, hdr.date);
    write32(out, hdr.crc32);
    write32(out, hdr.csize);
    write32(out, hdr.osize);
    write16(out, hdr.name.size());
    write16(out, hdr.extra.size());
    write16(out, hdr.comment.size());
    write16(out, 0);                        // disk number is always zero
    write16(out, hdr.internalAttrs);
    write32(out, hdr.externalAttrs);
    write32(out, hdr.offset);
    writeVector8(out, hdr.name);
    writeVector8(out, hdr.extra);
    writeVector8(out, hdr.comment);
}

// writes the central directory and other required records
void ZipWriter::finishArchive()
{
    writerData_t *d = static_cast<writerData_t *>(data);

    // write the file headers and compute the size of the central directory
    const uint32_t cdPos = ftell(d->out);

    uint32_t cdSize = 0;

    for (const auto &hdr : d->headers) {
        writeCDFileHeader(d->out, hdr);

        cdSize += 46;                   // size of the structure, DO NOT use sizeof() here!
        cdSize += hdr.name.size();
        cdSize += hdr.extra.size();
        cdSize += hdr.comment.size();
    }

    // @@@TODO: if Zip64, put here the Zip64 end of central directory record (APPNOTE.TXT chapter 4.3.14)
    // @@@TODO: if Zip64, put here the Zip64 central directory locator record (APPNOTE.TXT chapter 4.3.15)

    // central directory end record
    CDEnd_t end;

    end.id = kHeader_CDEnd;
    end.diskNum = 0;
    end.cdDisk = 0;
    end.filesOnDisk = d->headers.size();
    end.totalFiles = d->headers.size();
    end.size = cdSize;
    end.offset = cdPos;
    end.commentLen = d->archiveComment.length();

    fwrite(&end, sizeof(CDEnd_t), 1, d->out);

    // main archive comment
    if (!d->archiveComment.empty())
        fwrite(&d->archiveComment[0], 1, d->archiveComment.length(), d->out);
}

ZipWriter::ZipWriter() : compressionEnabled(true),
                         compressionLevel(Z_BEST_COMPRESSION)
{
    // setup the initial values
    writerData_t *d = new writerData_t;

    d->out = nullptr;
    d->errorCode = kError_None;

    data = d;
}

ZipWriter::~ZipWriter()
{
    finish();

    writerData_t *d = static_cast<writerData_t *>(data);

    delete d;
    data = nullptr;
}

/*
    Sets the main ZIP archive comment. The maximum length of the comment is 65536 characters.
    Longer comments will be truncated. Refer to the manual for notes about character encodings
    in comments.
*/
void ZipWriter::setArchiveComment(const string &comment)
{
    writerData_t *d = static_cast<writerData_t *>(data);

    string &c = d->archiveComment;

    c = comment;

    // truncate overly long comments
    if (c.length() > kMaxCommentLength)
        c = c.substr(0, kMaxCommentLength);
}

/*
    Returns the current archive main comment.
*/
const string &ZipWriter::getArchiveComment() const
{
    return static_cast<writerData_t *>(data)->archiveComment;
}

/*
    Returns true if an archive has been opened and it's ok to call addFile()/addDirectory().
*/
bool ZipWriter::ok() const
{
    return !!static_cast<writerData_t *>(data)->out;
}

/*
    Returns the current error code. Use errorString() to get textual error messages.
*/
int ZipWriter::getError() const
{
    return static_cast<writerData_t *>(data)->errorCode;
}

/*
    Converts the error code to a readable UTF-8 encoded error string. You can show this
    to the user, log it, etc. If no argument is given, the function returns the most
    recent error.
*/
const char *ZipWriter::errorString(const int error)
{
    const int code = (error == -1) ? getError() : error;

    switch (code) {
        case kError_None:
            return "No errors";

        case kError_ArchiveNotOpened:
            return "No ZIP file has been opened yet";

        case kError_ArchiveAlreadyOpen:
            return "A ZIP file has been already opened";

        case kError_CouldNotOpen:
            return "Could not open the specified file";

        case kError_EmptyName:
            return "A required filename was empty";

        case kError_UnicodeConversionFailed:
            return "UTF-8 to UCS-2 unicode conversion failed";

        case kError_DeflateInit2_Failed:
            return "deflateInit2() failed to initialize the stream";

        case kError_CouldNotReadFile:
            return "Could not read from the file";

        case kError_CouldNotWriteFile:
            return "Could not write to the file";

        case kError_CouldNotGetInfo:
            return "Could not get any information about the specified file or directory";

        default:
            return "Unknown error code passed to ZipWriter::errorString()";
    }
}

/*
    Returns the version and compilation date.
*/
const char *ZipWriter::getVersion()
{
    return "1.2 (2014-08-11)";
}

/*
    Returns the copyright string.
*/
const char *ZipWriter::getCopyright()
{
    return "Copyright © Jarmo Pietiläinen 2013-2014";
}

/*
    Opens a ZIP archive. After this, you can call addFile() and addDirectory() to add stuff
    into the archive. Once done, either call finish(), or just let the class go out of scope
    (it is automatically finished).

    Returns true if ok, false if failed.
*/
bool ZipWriter::open(const string &name)
{
    writerData_t *d = static_cast<writerData_t *>(data);

    if (d->out) {
        d->errorCode = kError_ArchiveAlreadyOpen;
        return false;
    }

    if (name.empty()) {
        d->errorCode = kError_EmptyName;
        return false;
    }

    // open the file
#ifdef ZIPWRITER_WINDOWS
    wstring wideName;

    if (!UTF8ToUCS2(name, wideName)) {
        d->errorCode = kError_UnicodeConversionFailed;
        return false;
    }

    d->out = _wfopen(wideName.c_str(), L"wb");
#else
    d->out = fopen(name.c_str(), "wb");
#endif

    if (!d->out) {
        d->errorCode = kError_CouldNotOpen;
        return false;
    }

    // okay, the file is open, reset the variables
    d->headers.clear();
    d->archiveComment = "";
    d->errorCode = kError_None;

#ifdef ZIPWRITER_WINDOWS
    // DOS/Windows (there's a value 10 for NTFS, but it is interpreted as "TOPS"
    // and that's not what we want)
    d->versionMadeBy = 0;
#else
    // Unix (I know, a gross generalization)
    d->versionMadeBy = 3;
#endif

    // set our version number (major*10+minor)
    d->versionMadeBy <<= 8;
    d->versionMadeBy |= 1 * 10 + 2;

    return true;
}

/*
    Finishes the archive writing. Writes out the central directory and closes the file.
    After this, you cannot add more files or directories into the archive without opening
    another one first.

    Returns true if OK, false if failed. The function fails if no archive is open.
*/
bool ZipWriter::finish()
{
    writerData_t *d = static_cast<writerData_t *>(data);

    if (!d->out) {
        d->errorCode = kError_ArchiveNotOpened;
        return false;
    }

    // we have an unfinished archive, finish it
    finishArchive();

    fclose(d->out);
    d->out = nullptr;

    return true;
}

/*
    Cancels the ongoing ZIP writing process. Central directory will not be written.
    If there's no ZIP open at the moment, this fails silently.
*/
void ZipWriter::cancel()
{
    writerData_t *d = static_cast<writerData_t *>(data);

    if (!d->out)
        return;

    fclose(d->out);
    d->out = nullptr;
}

/*
    Adds a file into the archive. You can pass in an optional file comment. The file
    names should contain their paths sans the "root" part. If 'internalName' is empty,
    then 'name' is used internally. This allows you to rename files on-the-fly.

    If the class member 'compressionEnabled' is false, the file will not be compressed.
    Otherwise it is compressed using the DEFLATE algorithm. You can set the compression
    level through the class member 'compressionLevel', which is between 0 and 9
    (or Z_NO_COMPRESSION, Z_BEST_SPEED, Z_BEST_COMPRESSION or Z_DEFAULT_COMPRESSION).

    Refer to the manual for a note about character encodings in comments.
*/

bool ZipWriter::addFile(const string &name, const string &internalName, const string &comment)
{
    writerData_t *d = static_cast<writerData_t *>(data);

    // ----------------------------------------------------------------------------------
    // initialize

    if (!d->out) {
        d->errorCode = kError_ArchiveNotOpened;
        return false;
    }

    if (name.empty()) {
        d->errorCode = kError_EmptyName;
        return false;
    }

    fileHeader_t hdr;

    if (internalName.empty())
        hdr.name.assign(name.begin(), name.end());
    else hdr.name.assign(internalName.begin(), internalName.end());

    forwardSlashes(hdr.name);

    hdr.comment.assign(comment.begin(), comment.end());

    // truncate overly long comments
    if (hdr.comment.size() > kMaxCommentLength)
        hdr.comment.resize(kMaxCommentLength);

    hdr.gpf |= kGPF_EFS;
    hdr.versionMadeBy = d->versionMadeBy;
    hdr.offset = ftell(d->out);

    // get external file information
#ifdef ZIPWRITER_WINDOWS
    wstring wideName;

    if (!UTF8ToUCS2(name, wideName)) {
        d->errorCode = kError_UnicodeConversionFailed;
        return false;
    }

    if (!getExternalFileInfo(wideName, hdr, false)) {
#else
    if (!getExternalFileInfo(name, hdr, false)) {
#endif
        d->errorCode = kError_CouldNotGetInfo;
        return false;
    }

    // ----------------------------------------------------------------------------------
    // open the file and copy/compress it

    file in;

#ifdef ZIPWRITER_WINDOWS
    if (!in.open(wideName.c_str(), L"rb")) {
#else
    if (!in.open(name.c_str(), "rb")) {
#endif
        d->errorCode = kError_CouldNotOpen;
        return false;
    }

    // leave room for the header (30 == size of the local file header minus the name/extra fields)
    for (size_t i = 0; i < hdr.name.size() + hdr.extra.size() + 30; i++)
        fputc(0, d->out);

    static const size_t kCopyBufferSize = 4096;

    uint8_t rbuffer[kCopyBufferSize];
    uint8_t cbuffer[kCopyBufferSize];

    // don't compress too small files
    // @@@TODO: need better heuristics here
    static const uint32_t kSmallEnoughFile = 16;

    const bool compress = (hdr.osize < kSmallEnoughFile) ? false : true;

    hdr.osize = 0;
    hdr.crc32 = crc32(0, Z_NULL, 0);

    if (compress && compressionEnabled) {
        // ------------------------------------------------------------------------------
        // compress using the DEFLATE algorithm

        // init the zlib stream
        z_stream_s stream;

        memset(&stream, 0, sizeof(z_stream_s));

        // use manual init, so zlib does not output its own initial bytes,
        // but gives us the actual raw stream
        int ret = deflateInit2(&stream, compressionLevel, Z_DEFLATED,
                               -MAX_WBITS, 8, Z_DEFAULT_STRATEGY);

        if (ret != Z_OK) {
            d->errorCode = kError_DeflateInit2_Failed;
            return false;
        }

        int flush = 0;

        do {
            // read in a block
            stream.avail_in = fread(rbuffer, 1, kCopyBufferSize, in);

            if (ferror(in)) {
                deflateEnd(&stream);
                d->errorCode = kError_CouldNotReadFile;
                return false;
            }

            hdr.osize += stream.avail_in;
            hdr.crc32 = crc32(hdr.crc32, rbuffer, stream.avail_in);

            flush = feof(in) ? Z_FINISH : Z_NO_FLUSH;
            stream.next_in = rbuffer;

            // compress and write
            do {
                stream.avail_out = kCopyBufferSize;
                stream.next_out = cbuffer;

                if (deflate(&stream, flush) == Z_STREAM_ERROR) {
                    deflateEnd(&stream);
                    d->errorCode = kError_CompressionFailed;
                    return false;
                }

                const size_t have = kCopyBufferSize - stream.avail_out;

                hdr.csize += have;

                if (fwrite(cbuffer, 1, have, d->out) != have || ferror(d->out)) {
                    deflateEnd(&stream);
                    d->errorCode = kError_CouldNotWriteFile;
                    return false;
                }

            } while (stream.avail_out == 0);
        } while (flush != Z_FINISH);

        deflateEnd(&stream);

        hdr.method = kCompression_Deflate;
        hdr.versionNeeded = kVersionNeeded_Default;
    } else {
        // ------------------------------------------------------------------------------
        // just copy the file as-is

        while (true) {
            const size_t n = fread(rbuffer, 1, kCopyBufferSize, in);

            if (n == 0)
                break;

            if (fwrite(rbuffer, 1, n, d->out) != n) {
                d->errorCode = kError_CouldNotWriteFile;
                return false;
            }

            hdr.crc32 = crc32(hdr.crc32, rbuffer, n);
            hdr.osize += n;
        }

        hdr.csize = hdr.osize;
        hdr.method = kCompression_None;
        hdr.versionNeeded = kVersionNeeded_Default;
    }

    in.close();

    // ----------------------------------------------------------------------------------
    // write the local header and finish the file

    const uint32_t pos = ftell(d->out);

    fseek(d->out, hdr.offset, SEEK_SET);
    writeLocalFileHeader(d->out, hdr);
    fseek(d->out, pos, SEEK_SET);

    d->headers.push_back(hdr);

    return true;
}

/*
    Adds a directory. You don't actually have to add directories manually; the file names
    in ZIP archives already contain directory names. It's the extractor's job to analyze
    which directories are contained in those names and then recreate them if needed. Even
    the venerable DOS PKZip 2.04g can recreate the directory structure just fine without
    directory entries (I've tested this).

    However, if you want to retain original directory timestamps and/or attributes, then
    you need to add directories. This operation could be automated by extracting the
    directory names from every added file and maintaining a tree internally. This may be
    implemented in the future versions, but I won't promise anything. For now, you have
    to do things manually.

    The name is always stored UTF-8 encoded.
*/
bool ZipWriter::addDirectory(const string &name, const string &internalName)
{
    writerData_t *d = static_cast<writerData_t *>(data);

    if (!d->out) {
        d->errorCode = kError_ArchiveNotOpened;
        return false;
    }

    if (name.empty()) {
        d->errorCode = kError_EmptyName;
        return false;
    }

    fileHeader_t hdr;

    if (internalName.empty())
        hdr.name.assign(name.begin(), name.end());
    else hdr.name.assign(internalName.begin(), internalName.end());

    forwardSlashes(hdr.name);

    // append a slash at the end if needed
    if (*hdr.name.rbegin() != '/')
        hdr.name.push_back('/');

    hdr.gpf |= kGPF_EFS;
    hdr.versionNeeded = kVersionNeeded_Directory;
    hdr.versionMadeBy = d->versionMadeBy;
    hdr.offset = ftell(d->out);

    // get external directory information
#ifdef ZIPWRITER_WINDOWS
    wstring wideName;

    if (!UTF8ToUCS2(name, wideName)) {
        d->errorCode = kError_UnicodeConversionFailed;
        return false;
    }

    if (!getExternalFileInfo(wideName, hdr, true)) {
#else
    if (!getExternalFileInfo(name, hdr, true)) {
#endif
        d->errorCode = kError_CouldNotGetInfo;
        return false;
    }

    // nothing needs to be written besides the local header, so we're good
    writeLocalFileHeader(d->out, hdr);
    d->headers.push_back(hdr);

    return true;
}
