/*
    *************************************************************************************

    ZipWriter v1.2
    (c) Jarmo Pietiläinen 2013-2014
    http://z0b.kapsi.fi

    This software is provided 'as-is', without any express or implied warranty.
    In no event will the authors be held liable for any damages arising from the
    use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it freely,
    subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not
         be misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    *************************************************************************************
*/

#ifndef ZIPWRITER_H
#define ZIPWRITER_H

#include <string>

class ZipWriter {
public:
    // error codes
    enum
    {
        kError_None,
        kError_ArchiveNotOpened,
        kError_ArchiveAlreadyOpen,
        kError_CouldNotOpen,
        kError_EmptyName,
        kError_UnicodeConversionFailed,
        kError_DeflateInit2_Failed,
        kError_CompressionFailed,
        kError_CouldNotReadFile,
        kError_CouldNotWriteFile,
        kError_CouldNotGetInfo,
    };

    ZipWriter();
    virtual ~ZipWriter();

    // Opens an archive for writing. Will fail (kError_ArchiveAlreadyOpen) if this
    // class instance already has an archive open. Close the previous archive first.
    bool open(const std::string &name);

    // Adds a file into the archive. 'name' is the real name of the file you want
    // to add, and 'internalName' is the name the file gets inside the archive. If
    // 'internalName' is empty, 'name' will be used. 'fileComment' is a comment for
    // the file. See the manual for notes about character encodings.
    bool addFile(const std::string &name,
                 const std::string &internalName = "",
                 const std::string &fileComment = "");

    // Adds a directory into the archive. You only have to add directoris if you want
    // to preserve their timestamps and attributes. All ZIP extractors I've tested can
    // recreate the directory structure from the filenames, but that won't preserve the
    // timestamps or attributes.
    bool addDirectory(const std::string &name,
                      const std::string &internalName = "");

    // Call this to finish the archive and close it. To add more files/directories
    // after this, call open(). You don't have to call this, because the destructor
    // calls this, but it doesn't hurt to make sure.
    bool finish();

    // Call this to cancel the archive writing and close it without finishing it;
    // the files/directories already added remain in the archive, but the archive
    // will not have its central directory written out. Tools exist for "repairing"
    // these unfinished archives.
    void cancel();

    // returns true if an archive has been opened and you can add files and directories
    // into the archive
    bool ok() const;

    // Set/get the main archive comment. You can call these any time between
    // open() and finish(). See the manual for notes about character encodings.
    void setArchiveComment(const std::string &comment);
    const std::string &getArchiveComment() const;

    // error handling
    const char *errorString(const int error = -1);
    int getError() const;

    // Compression mode and level; you can change these at anytime. Changing these
    // affects the next file you add and everything after it until you change these
    // again. These won't get reset if you close the archive. 'compressionLevel' is
    // one of the Z_* compression level constants defined in zlib.h. By default,
    // compression is enabled and level is Z_BEST_COMPRESSION.
    bool compressionEnabled;
    int compressionLevel;

    // UTF-8 encoded
    static const char *getVersion();
    static const char *getCopyright();

    // no copying or assignments (you can't sensibly share file writing between
    // multiple copies of a single class instance)
    ZipWriter(const ZipWriter &writer) = delete;
    ZipWriter &operator = (const ZipWriter &writer) = delete;
    ZipWriter *operator = (const ZipWriter *writer) = delete;

protected:
    // opaque pointer to the internal state, stored in a void pointer so I don't
    // have to expose the gritty details to the outside world
    void *data;

    void finishArchive();
};

#endif
